@@ -6,71 +6,69 @@
 //
 // You should have received a copy of the GNU General Public License
 // along with Blades.  If not, see <http://www.gnu.org/licenses/>
-
 use crate::config::Config;
-use crate::error::Result;
-use crate::sources::{Source, Sources};
+use crate::sources::{Parser, Source, Sources};
 use crate::tasks::render;
 use crate::taxonomies::{Classification, Taxonomies};
 use crate::types::{Ancestors, Any, DateTime, HashMap, MutSet, Templates};
 
+use arrayvec::ArrayVec;
 use beef::lean::Cow;
 use ramhorns::{encoding::Encoder, traits::ContentSequence, Content, Section, Template};
-use serde::Deserialize;
+use serde::{Deserialize, Serialize};
 
 use std::cmp::{min, Reverse};
 use std::fs::create_dir_all;
+use std::io;
 use std::num::NonZeroUsize;
 use std::ops::Range;
 use std::path::{is_separator, Path, PathBuf};
 
 /// All the information regarding one page
-#[derive(Content, Deserialize)]
+#[derive(Content, Default, Deserialize, Serialize)]
 pub struct Page<'p> {
-    #[serde(borrow, default)]
+    #[serde(borrow, default, skip_serializing_if = "str::is_empty")]
     title: Cow<'p, str>,
-    #[serde(borrow, default)]
+    #[serde(borrow, default, skip_serializing_if = "str::is_empty")]
     slug: Cow<'p, str>,
-    #[serde(borrow, default)]
+    #[serde(borrow, default, skip_serializing_if = "is_ancestors_empty")]
     path: Ancestors<'p>,
-    #[serde(default)]
+    #[serde(default, skip_serializing_if = "is_cow_empty")]
     alternative_paths: Cow<'p, [&'p str]>,
-    #[serde(default)]
+    #[serde(default, skip_serializing_if = "equal_zero")]
     #[ramhorns(skip)]
     pub(crate) weight: i64,
-    #[serde(borrow, default)]
+    #[serde(borrow, default, skip_serializing_if = "str::is_empty")]
     #[ramhorns(skip)]
     template: Cow<'p, str>,
-    #[serde(borrow, default = "default_page")]
+    #[serde(borrow, default = "default_page", skip_serializing_if = "eq_default_page")]
     #[ramhorns(skip)]
     page_template: Cow<'p, str>,
-    #[serde(borrow, default = "default_section")]
+    #[serde(borrow, default = "default_section", skip_serializing_if = "eq_default_section")]
     #[ramhorns(skip)]
     section_template: Cow<'p, str>,
-    #[serde(borrow, default = "default_gallery")]
+    #[serde(borrow, default = "default_gallery", skip_serializing_if = "eq_default_gallery")]
     #[ramhorns(skip)]
     gallery_template: Cow<'p, str>,
+    #[serde(skip_serializing_if = "Option::is_none")]
     #[ramhorns(skip)]
     paginate_by: Option<NonZeroUsize>,
-    #[serde(default)]
-    #[ramhorns(skip)]
-    pictures: Cow<'p, [Picture<'p>]>,
-    #[serde(borrow, default)]
+    #[serde(borrow, default, skip_serializing_if = "str::is_empty")]
     image: Cow<'p, str>,
-    #[serde(borrow, default)]
+    #[serde(borrow, default, skip_serializing_if = "str::is_empty")]
     summary: Cow<'p, str>,
-    #[serde(borrow, default)]
+    #[serde(borrow, default, skip_serializing_if = "str::is_empty")]
     #[md]
     content: Cow<'p, str>,
 
     pub(crate) date: Option<DateTime>,
 
-    #[serde(default)]
+    #[serde(default, skip_serializing_if = "std::ops::Not::not")]
     #[ramhorns(skip)]
     sort_by_weight: bool,
     #[serde(skip)]
     is_section: bool,
-    #[serde(default)]
+    #[serde(default, skip_serializing_if = "std::ops::Not::not")]
     hidden: bool,
 
     #[serde(skip, default = "default_range")]
@@ -92,9 +90,13 @@ pub struct Page<'p> {
     #[serde(skip, default = "default_priority")]
     priority: f32,
 
-    #[serde(default)]
+    #[serde(default, skip_serializing_if = "ArrayVec::is_empty")]
+    #[ramhorns(skip)]
+    pictures: ArrayVec<Picture<'p>, 4>,
+
+    #[serde(default, skip_serializing_if = "HashMap::is_empty")]
     pub(crate) taxonomies: Taxonomies<'p>,
-    #[serde(default)]
+    #[serde(default, skip_serializing_if = "HashMap::is_empty")]
     extra: HashMap<&'p str, Any<'p>>,
 
     /// A unique number to determine whether this is the active page
@@ -103,7 +105,7 @@ pub struct Page<'p> {
     id: usize,
 }
 
-#[derive(Clone, Content, Deserialize)]
+#[derive(Clone, Content, Deserialize, Serialize)]
 pub(crate) struct Picture<'p> {
     #[serde(borrow, default)]
     alt: Cow<'p, str>,
@@ -220,7 +222,7 @@ pub(crate) trait Paginate: Content + Sized {
         path: &mut PathBuf,
         tpl: &Template,
         rendered: &MutSet,
-    ) -> Result<()> {
+    ) -> Result<(), ramhorns::Error> {
         let count = last - first;
         let by = min(by, count);
         let len = count / by + ((count % by != 0) as usize);
@@ -245,9 +247,16 @@ pub(crate) trait Paginate: Content + Sized {
 impl<'p> Page<'p> {
     /// Construct a new page from the source.
     #[inline]
-    pub fn new(source: &'p Source, data: &'p Sources, config: &Config) -> Result<Self> {
-        let mut page: Page = toml::from_slice(&data.data[source.source.clone()])
-            .map_err(|e| (e, source.path.clone()))?;
+    pub fn new<P: Parser>(
+        source: &'p Source<P>,
+        data: &'p Sources<P>,
+        config: &Config,
+    ) -> Result<Self, (P::Error, Box<str>)> {
+        let path = std::str::from_utf8(&data.data[source.path.clone()]).unwrap();
+        let mut page = source
+            .format
+            .parse(&data.data[source.source.clone()])
+            .map_err(|e| (e, path.into()))?;
 
         let is_section = source.is_section;
         page.is_section = is_section;
@@ -258,14 +267,13 @@ impl<'p> Page<'p> {
             page.date = page.date.or_else(|| source.date.map(|d| d.into()));
         }
 
-        let path = &source.path;
+        // The path is already ensured to be valid UTF-8
         let path = path
             .strip_prefix(config.content_dir.as_ref())
             .unwrap_or(path);
         let path = path.strip_prefix(is_separator).unwrap_or(path);
         if is_section || page.slug.is_empty() || page.slug.contains(is_separator) {
             let slug = path.rsplit(is_separator).next().unwrap_or_default();
-            let slug = slug.strip_suffix(".toml").unwrap_or(slug);
             page.slug = Cow::const_str(slug);
         }
         let page_path = page.path.as_ref();
@@ -279,10 +287,16 @@ impl<'p> Page<'p> {
         Ok(page)
     }
 
+    /// Replace the page content with `content`.
+    pub fn with_content(mut self, content: Cow<'p, str>) -> Self {
+        self.content = content;
+        self
+    }
+
     /// Appropriately sort the given vector of pages and create all the directories where
     /// they will be rendered to.
     #[inline]
-    pub fn prepare(mut pages: Vec<Self>, config: &Config) -> Result<Vec<Self>> {
+    pub fn prepare(mut pages: Vec<Self>, config: &Config) -> Result<Box<[Self]>, io::Error> {
         let output_dir = Path::new(config.output_dir.as_ref());
         for i in 0..pages.len() {
             let page = &pages[i];
@@ -329,7 +343,7 @@ impl<'p> Page<'p> {
             pages[i].id = i;
         }
 
-        Ok(pages)
+        Ok(pages.into())
     }
 
     /// Get a reference of the page, in context of its subpages and subsections.
@@ -383,7 +397,7 @@ impl<'p> Page<'p> {
         config: &Config<'p>,
         classification: &Classification<'p, '_>,
         rendered: &MutSet,
-    ) -> Result {
+    ) -> Result<(), ramhorns::Error> {
         let output_dir = Path::new(config.output_dir.as_ref());
         let mut output = output_dir.join(self.path.as_ref());
         output.push(self.slug.as_ref());
@@ -637,3 +651,33 @@ const fn default_page() -> Cow<'static, str> {
 const fn default_section() -> Cow<'static, str> {
     Cow::const_str("section.html")
 }
+
+#[inline]
+fn eq_default_gallery(c: &str) -> bool {
+    c == "gallery.html"
+}
+
+#[inline]
+fn eq_default_page(c: &str) -> bool {
+    c == "page.html"
+}
+
+#[inline]
+fn eq_default_section(c: &str) -> bool {
+    c == "section.html"
+}
+
+#[inline]
+const fn equal_zero(i: &i64) -> bool {
+    *i == 0
+}
+
+#[inline]
+fn is_cow_empty<T: Clone>(s: &Cow<[T]>) -> bool {
+    s.is_empty()
+}
+
+#[inline]
+fn is_ancestors_empty(s: &Ancestors) -> bool {
+    s.0.is_empty()
+}
